module Data.Decimal.Frisby
    ( decimal
    ) 
    where

import Control.Monad(ap)
import Data.Char
import Text.Parsers.Frisby
import Data.Decimal.Representation
import Data.Decimal.Print

(<:>) :: P s a -> P s [a] -> P s [a]
p <:> ps = p <> ps ## uncurry (:)

-- * Decimal

type Digits = [Char]

digit :: P s Char
digit = oneOf ['0'..'9']

digits :: P s Digits
digits = many1 digit

sign :: P s Sign
sign =
    (char '+' ##> Positive) // (char '-' ##> Negative) // unit Positive

nan :: P s Decimal
nan = 
    (sign <> (itext "NaN"  ->> option Nothing (many1 digit ## Just . read)) ## uncurry QuietNaN) //
    (sign <> (itext "sNaN" ->> option Nothing (many1 digit ## Just . read)) ## uncurry SignalNaN)

infinity :: P s Decimal
infinity =
    sign <<- (itext "Inf" ->> optional (itext "inity")) ## Infinity

decimalPart :: P s (Digits, Digits)
decimalPart =
    (many digit <> (char '.' ->> digits)) // (digits <> unit "")

exponentPart :: P s Integer
exponentPart =
    sign <> digits ## \(s,d) ->
        case s of 
          Positive -> read d
          Negative -> (-(read  d))

indicator :: P s Char
indicator = char 'E' // char 'e'

normalDecimal =
    sign <> (decimalPart <> (indicator ->> exponentPart // unit 0)) ## \(sign, ((beforePoint, afterPoint), ep)) ->
        let exp  = ep - (fromIntegral (length afterPoint))
            coef = read (beforePoint ++ afterPoint)
        in (Decimal sign coef exp)

-- frisby instance of decimal
decimal :: P s Decimal
decimal = choice [nan, infinity, normalDecimal]

-- (digits <> unit "")

-- | match some text case-insensitively
itext :: String -> P s String
itext (x:xs) = fmap ( \ (c,cs) -> c:cs) $ (char (toUpper x) // char (toLower x)) <> itext xs
itext [] = unit []

-- * Test Stuff
  
test :: String -> Decimal
test s = runPeg (newRule decimal) s        

runTests :: IO ()
runTests = mapM_ runTest testValues
    where
      runTest (s,d) = 
          let d' = runPeg (newRule decimal) s
          in if (show d') == (show d)
             then putStrLn $ "ok: " ++ s ++ " " ++ show d
             else putStrLn $ "fail: " ++ s ++ " " ++ show d ++ " " ++ show d'

testValues :: [(String, Decimal)]
testValues =
    [ ("0", Decimal Positive 0 0)
    , ("0.00", Decimal Positive 0 (-2))
    , ("123", Decimal Positive 123 0)
    , ("-123", Decimal Negative 123 0)
    , ("1.23E3", Decimal Positive 123 1)
    , ("1.23E+3", Decimal Positive 123 1)
    , ("12.3E+7", Decimal Positive 123 6)
    , ("12.0", Decimal Positive 120 (-1))
    , ("12.3", Decimal Positive 123 (-1))
    , ("0.00123", Decimal Positive 123 (-5))
    , ("-1.23E-12", Decimal Negative 123 (-14))
    , ("1234.5E-4", Decimal Positive 12345 (-5))
    , ("-0", Decimal Negative 0 0)
    , ("-0.00", Decimal Negative 0 (-2))
    , ("0E+7", Decimal Positive 0 7)
    , ("-0E-7", Decimal Negative 0 (-7))
    , ("inf", Infinity Positive)
    , ("+inFiniTy", Infinity Positive)
    , ("-Infinity", Infinity Negative)
    , ("NaN", QuietNaN Positive Nothing)
    , ("-NAN", QuietNaN Negative Nothing)
    , ("SNaN", SignalNaN Positive Nothing)
    ]

