module Data.Decimal.Operations where

import Data.Ratio
import Debug.Trace
import Data.Decimal.Representation
import Data.Decimal.Print

instance Eq Decimal where
    a == b = 
	let ((Decimal sa coefa expa), (Decimal sb coefb expb)) = alignDecimal a b
	    in
	    (sa == sb && coefa == coefb && expa == expb)

instance Ord Decimal where
    compare a b = compareDecimal a b

instance Num Decimal where
    a + b    = addDecimal a b
    a - b    = addDecimal a (negate b)
    a * b    = multiplyDecimal a b
    negate a = negateDecimal a
    abs a    = absDecimal a
    signum a = signumDecimal a
    fromInteger i = fromIntegerDecimal i

instance Fractional Decimal where
    a / b    = divideDecimal a b
    fromRational = fromRationalDecimal

signXor :: Sign -> Sign -> Sign
signXor Positive Positive = Positive
signXor Positive Negative = Negative
signXor Negative Positive = Negative
signXor Negative Negative = Positive

alignDecimal :: Decimal -> Decimal -> (Decimal, Decimal)
alignDecimal a@(Decimal sa coefa expa) b@(Decimal sb coefb expb)
    | expa == expb =
	(a,b)
    | expa >  expb =
	((Decimal sa (coefa * (10 ^ (expa - expb))) expb),
	 (Decimal sb coefb expb))
    | expb >  expa =
	((Decimal sa coefa expa),
	 (Decimal sb (coefb * (10 ^ (expb - expa))) expa))

-- is there a better way?
lastIsOdd :: Decimal -> Bool
lastIsOdd (Decimal _ coef _) =
    let lastDigit = last (show coef) in
	odd (read [lastDigit])

lastIsEven :: Decimal -> Bool
lastIsEven = not . lastIsOdd

data RoundHint
    = RoundZero
    | RoundBelow
    | RoundOneHalf
    | RoundAbove
      deriving (Ord, Eq, Show)

roundRemainder :: Integer -> Integer -> RoundHint
roundRemainder 0 divisor = RoundZero
roundRemainder dividen divisor =
    case compare (dividen % divisor) (1%2) of
	 LT -> RoundBelow
	 EQ -> RoundOneHalf
	 GT -> RoundAbove

roundDecimal :: Rounding -> Decimal -> RoundHint -> Decimal
roundDecimal RoundDown d _ = d
roundDecimal RoundHalfUp d@(Decimal s coef exp) roundHint
    | roundHint >= RoundOneHalf = (Decimal s (coef+1) exp)
    | otherwise = d
roundDecimal RoundHalfEven d@(Decimal s coef exp) roundHint
    | roundHint > RoundOneHalf = (Decimal s (coef + 1) exp)
    | roundHint == RoundOneHalf && (lastIsEven d) = (Decimal s (coef + 1) exp) -- FIXME: this may be wrong, see pep327
    | otherwise = d
roundDecimal RoundCeiling d@(Decimal s coef exp) roundHint
    | s == Negative || roundHint == RoundZero = d
    | otherwise = (Decimal s (coef + 1) exp)
roundDecimal RoundFloor d@(Decimal s coef exp) roundHint
    | s == Positive || roundHint == RoundZero = d
    | otherwise = (Decimal s (coef + 1) exp)
roundDecimal RoundHalfDown d@(Decimal s coef exp) roundHint
    | roundHint > RoundOneHalf = (Decimal s (coef+1) exp)
    | otherwise = d
roundDecimal RoundUp d@(Decimal s coef exp) roundHint
    | roundHint == RoundZero = d
    | otherwise = (Decimal s (coef + 1) exp)

absDecimal :: Decimal -> Decimal
absDecimal (Decimal _ coef exp) = (Decimal Positive coef exp)

-- |Add special value handlers
addDecimal :: Decimal -> Decimal -> Decimal
addDecimal a@(Decimal _ _ _) b@(Decimal _ _ _) =
    let ((Decimal sa coefa expa),(Decimal sb coefb expb)) = alignDecimal a b
	newCoef =
	    case (sa == sb, coefa > coefb) of
		 (True,  _)     -> coefa + coefb
		 (False, True)  -> coefa - coefb
		 (False, False) -> coefb - coefa
	newSign = -- does not handle 'FIXME: signs of operands different and rounding is round-floor'
	    case (newCoef,sa,sb,coefa > coefb) of
		 (0, Negative, Negative,_) -> Negative
		 (0, _, _,_) -> Positive
		 (_,_,_,True) -> sa
		 (_,_,_,False) -> sb
	in
	(Decimal newSign newCoef expa)

negateDecimal :: Decimal -> Decimal
negateDecimal (Decimal Positive coef exp) = (Decimal Negative coef exp)
negateDecimal (Decimal Negative coef exp) = (Decimal Positive coef exp)

compareDecimal :: Decimal -> Decimal -> Ordering
compareDecimal a@(Decimal sa coefa expa) b@(Decimal sb coefb expb) =
    if sa /= sb
       then compare (quick a) (quick b)
       else case (a - b) of
		 (Decimal _ 0 _) -> EQ
		 (Decimal Positive _ _) -> GT
		 (Decimal Negative _ _) -> LT
    where quick :: Decimal -> Int
	  quick (Decimal _ 0 _) = 0
	  quick (Decimal Positive _ _) = 1
	  quick (Decimal Negative _ _) = (-1)

multiplyDecimal :: Decimal -> Decimal -> Decimal
multiplyDecimal (Decimal sa coefa expa) (Decimal sb ceofb expb) =
    (Decimal (signXor sa sb) (coefa * ceofb) (expa + expb))

signumDecimal :: Decimal -> Decimal
signumDecimal (Decimal _ 0 _)        = (Decimal Positive 0 0)
signumDecimal (Decimal Positive _ _) = (Decimal Positive 1 0)
signumDecimal (Decimal Negative _ _) = (Decimal Negative 1 0)

fromIntegerDecimal :: Integer -> Decimal
fromIntegerDecimal i
    | i < 0     = (Decimal Negative (abs i) 0)
    | otherwise = (Decimal Positive i 0)
	  
divideDecimal :: Decimal -> Decimal -> Decimal
divideDecimal a@(Decimal sa 0 _) b@(Decimal sb 0 _) = 
    SignalNaN Positive Nothing -- FIXME
divideDecimal a@(Decimal sa _ _) b@(Decimal sb 0 _) =
    (Infinity (signXor sa sb))
-- Add rounding, etc
divideDecimal a@(Decimal sa coefa expa) b@(Decimal sb coefb expb) =
    let (adjust, coefa',coefb') = adjustCoefs 0 coefa coefb
	(coef', roundHint, adjust') = divideCoefs adjust 0 coefa' coefb'
	exp' = expa - (expb + adjust')
	in
	roundDecimal RoundHalfEven (Decimal (signXor sa sb) coef' exp') roundHint
    where adjustCoefs adjust coefa coefb
	      | coefa < coefb = 
		  adjustCoefs (adjust + 1) (coefa * 10) coefb
	      | coefa >= (coefb * 10) =
		  adjustCoefs (adjust - 1) coefa (coefb * 10)
	      | otherwise = (adjust, coefa, coefb)
	  divideCoefs adjust result coefa coefb
	      | coefb <=  coefa =
		  let result' = result + 1
		      coefa'  = coefa - coefb
		      in
		      divideCoefs adjust result' coefa' coefb
	      | ((coefa == 0 && adjust >= 0) ||
		 (length (show result) >= 9)) = --FIXME (not fixed length)
		         (result, roundRemainder coefa coefb, adjust)
	      | otherwise =
		  divideCoefs (adjust + 1) (result * 10) (coefa * 10) coefb
		
fromRationalDecimal :: Rational -> Decimal
fromRationalDecimal r =
    let n = fromIntegral (numerator r)
	d = fromIntegral (denominator r)
	in (n / d)
