module Data.Decimal.Parsec 
    ( decimal
    ) where

-- only export decimal ?

import Data.Char
import Data.Decimal.Representation
import Text.ParserCombinators.Parsec

type DecimalParser st a = CharParser st a

type Digits = [Char]

decimal :: CharParser st Decimal
decimal =
    do s <- sign <|> (return Positive)
       ((nan s) 
	<|>
	(infinity s)
	<|>
	do (beforePoint, afterPoint) <- decimalPart 
	   ep <- (do indicator 
		     exponentPart
		  <|>
		  return 0)
	   let exp  = ep - (fromIntegral (length afterPoint))
 	       coef = read (beforePoint ++ afterPoint)
	   return (Decimal s coef exp))

sign :: CharParser st Sign
sign = 
    do char '+'
       return Positive
    <|>
    do char '-'
       return Negative

indicator :: CharParser st Char
indicator = 
    char 'e' <|> char 'E'

digits :: CharParser st Digits
digits = 
    many1 digit

decimalPart :: CharParser st (Digits,Digits)
decimalPart =
    try (do d1 <- digits
	    char '.'
	    d2 <- digits <|> return ""
	    return (d1,d2))
    <|>
    do char '.'
       d2 <- digits
       return ("",d2)
    <|>
    do d1 <- digits
       return (d1,"")

exponentPart :: CharParser st Integer
exponentPart =
    do s <- sign <|> (return Positive)
       d <- digits
       case s of
	    Positive -> return   (read d)
	    Negative -> return (-(read d))

infinity :: Sign -> CharParser st Decimal
infinity s =
    do istring "Inf" 
       istring "inity" <|> return ""
       return (Infinity s)

nan :: Sign -> CharParser st Decimal
nan s =
    do istring "NaN"
       d <- maybep digits
       return (QuietNaN s (maybe Nothing read d))
    <|>
    do istring "sNaN"       
       d <- maybep digits
       return (SignalNaN s (maybe Nothing read d))

istring :: String -> CharParser a String
istring str =
    mapM oneOf (map (\c -> [toUpper c, toLower c]) str)

maybep :: GenParser tok st a -> GenParser tok st (Maybe a)
maybep p = 
    (p >>= (return . Just))
    <|>
    (return Nothing)

-- * Test code (add more)

checkValue :: (String, Decimal) -> Bool
checkValue (str, dec) =
    case parse decimal str str of
	 Left e     -> False
--	 Right dec' -> (dec == dec')
	 Right _ -> True

testParser = 
    all checkValue testValues


testValues :: [(String, Decimal)]
testValues =
    [ ("0", Decimal Positive 0 0)
    , ("0.00", Decimal Positive 0 (-2))
    , ("123", Decimal Positive 123 0)
    , ("-123", Decimal Negative 123 0)
    , ("1.23E3", Decimal Positive 123 1)
    , ("1.23E+3", Decimal Positive 123 1)
    , ("12.3E+7", Decimal Positive 123 6)
    , ("12.0", Decimal Positive 120 (-1))
    , ("12.3", Decimal Positive 123 (-1))
    , ("0.00123", Decimal Positive 123 (-5))
    , ("-1.23E-12", Decimal Negative 123 (-14))
    , ("1234.5E-4", Decimal Positive 12345 (-5))
    , ("-0", Decimal Negative 0 0)
    , ("-0.00", Decimal Negative 0 (-2))
    , ("0E+7", Decimal Positive 0 7)
    , ("-0E-7", Decimal Negative 0 (-7))
    , ("inf", Infinity Positive)
    , ("+inFiniTy", Infinity Positive)
    , ("-Infinity", Infinity Negative)
    , ("NaN", QuietNaN Positive Nothing)
    , ("-NAN", QuietNaN Negative Nothing)
    , ("SNaN", SignalNaN Positive Nothing)
    ]

